﻿namespace pipeline
{
    using GrymCore;
    using stdole;
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Globalization;
    using System.Linq;
    using System.Runtime.InteropServices;
    using System.Text;
    using System.Xml.Linq;

    /*
     * Базовый класс для абстракции поставщика данных,
     * плюс вспомогательные структуры.
     */
    internal abstract class Data
    {
        public static readonly NumberFormatInfo NUMBER_FORMAT = _init_NUMBER_FORMAT();
        private static NumberFormatInfo _init_NUMBER_FORMAT()
        {
            var rv = (NumberFormatInfo)NumberFormatInfo.CurrentInfo.Clone();
            rv.NumberDecimalSeparator = ".";
            return rv;
        }

        public struct Point
        {
            public IMapPoint Local;
            public double Latitude;
            public double Longitude;

            public static Point Create(PluginEngine engine, IMapPoint local)
            {
                Debug.Assert((engine != null) && (local != null));

                var rv = new Point();
                var geo = engine.CoordinateTransformation.LocalToGeo(local);
                rv.Local = local;
                rv.Latitude = geo.X;
                rv.Longitude = geo.Y;

                return rv;
            }

            public static Point Create(PluginEngine engine, double latitude, double longitude)
            {
                var geo = engine.Factory.CreateMapPoint(latitude, longitude);
                return new Point()
                {
                    Latitude = latitude,
                    Longitude = longitude,
                    Local = engine.CoordinateTransformation.GeoToLocal(geo) 
                };
            }

            public bool InRect(double left, double top, double right, double bottom)
            {
                return ((Local.X < right) && (Local.Y < bottom) && (Local.X > left) && (Local.Y > top));
            }

            public bool InRect(IMapPoint min, IMapPoint max)
            {
                return InRect(min.X, min.Y, max.X, max.Y);
            }

            public bool InRect(IMapRect rect)
            {
                return InRect(rect.Min, rect.Max);
            }
        }

        public struct Label
        {
            public string Text;
            public string ShapeInfo;

            public static Label Create(string text)
            {
                var rv = new Label();
                rv.Text = string.Empty;
                rv.ShapeInfo = null;

                return rv;
            }
        }

        public struct Style
        {
            public enum Pen
            {
                Solid = 0,
                Dash = 1,
                Dot = 2,
                DashDot = 3,
                DashDotDot = 4
            }

            public string Id;
            public int LineColor;
            public int LineWidth;
            public Pen LinePen;

            public string IconPath;

            public bool LabelVisible;
            public int LabelColor;
            public IFont LabelFont;
            public int LabelStyle;

            public static Style Create(string icon, Pen pen)
            {
                var rv = new Style();
                rv.IconPath = icon;
                rv.LinePen = pen;

                return rv;
            }
        }

        private XDocument _xml;
        private PluginEngine _engine;
        private string[] _contentIds;
        private Dictionary<string, XElement> _content = new Dictionary<string,XElement>();

        public Data(PluginEngine engine)
        {
            Debug.Assert(engine != null);

            _engine = engine;
            _xml = XDocument.Load(_engine.KmlPath);
            _Init();
            Debug.Print(_xml.Document.ToString());
        }

        public abstract Point[] GetCoordinates(XElement node);
        public abstract Label GetLabel(XElement node);
        public abstract Style GetStyle(XElement node);
        public abstract void SetStyle(XElement node, string style);

        public virtual string[] ContentIds
        {
            get
            {
                if (_contentIds == null)
                {
                    _contentIds = Find()
                        .Where(x => GetId(x) != null)
                        .Select(x => GetId(x).Value)
                        .ToArray();
                }

                return _contentIds;
            }
        }

        public static int GetIntValue(XAttribute attr)
        {
            Debug.Assert((attr != null) && (attr.Value != null) && (attr.Value.Length != 0));
            int rv;
            return int.TryParse(attr.Value, out rv) ? rv : -1;
        }

        public static int GetIntValue(XElement node)
        {
            Debug.Assert((node != null) && (node.Value != null) && (node.Value.Length != 0));
            int rv;
            return int.TryParse(node.Value, out rv) ? rv : -1;

        }

        public static XAttribute GetId(XElement node)
        {
            Debug.Assert(node != null);

            return node.Attribute(XName.Get("id"));
        }

        public PluginEngine Engine
        {
            get { return _engine; }
        }

        public XElement Root
        {
            get
            {
                return _xml.Root;
            }
        }

        public void Load(string xml)
        {
            xml = xml.Trim();
            _xml = xml.Length == 0 ? XDocument.Load(_engine.KmlPath) : XDocument.Parse(xml); 
            _Init();
            _engine.Map.Invalidate(true);

            Debug.Print(_xml.Document.ToString());
        }

        public void Save()
        {
            _xml.Save(_engine.KmlPath);
        }

        public bool Exists(string id)
        {
            return Find(id) != null;
        }

        public IEnumerable<XElement> Find()
        {
            return Root.DescendantNodes().OfType<XElement>();
        }

        public IEnumerable<XElement> Find(string[] ids)
        {
            Debug.Assert(ids != null);

            if (ids.Equals(ContentIds))
                return _content.Values;


            var rv = new List<XElement>(ids.Length);
            if (ids.Length == 0)
                return rv;

            foreach (var id in ids)
            {
                var node = Find(id);
                if (node != null)
                    rv.Add(node);
            }

            return rv;

        }

        public IEnumerable<XElement> Find(IMapRect area)
        {
            var rv = new List<XElement>(ContentIds.Length);

            foreach (var node in Find(ContentIds))
            {
                foreach (var point in GetCoordinates(node))
                {
                    if (point.InRect(area.MinX, area.MinY, area.MaxX, area.MaxY))
                    {
                        rv.Add(node);
                        break;
                    }
                }
            }

            return rv;
        }

        public XElement Find(string id)
        {
            Debug.Assert(id != null);

            return _content.ContainsKey(id) ? _content[id] : null;
        }

        public Point MakeCoordinate(IMapPoint point)
        {
            return Data.Point.Create(Engine, point);
        }

        private void _Init()
        {
            _content.Clear();
            if (ContentIds.Length == 0)
                return;
            
            XAttribute attr = null;
            var search = new HashSet<string>(ContentIds);
            foreach (var node in Find().Where(x => ((attr = GetId(x)) != null) && search.Contains(attr.Value)))
                _content.Add(attr.Value, node);

            Debug.Assert(_content.First().Key != null);
        }
    }
}
