﻿using stdole;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Xml.Linq;


namespace pipeline
{
    internal sealed class KmlData : Data
    {
        public KmlData(PluginEngine engine) : base(engine)
        {
            
        }

        private XElement _FindElement(string name, XElement node = null)
        {
            Debug.Assert(name.Length != 0);

            if (node == null)
                node = Root.Element(XName.Get("Document", Root.Name.NamespaceName));
                    
            return node.Element(XName.Get(name, node.Name.NamespaceName));
        }

        private XElement _FindStyle(string id)
        {
            Debug.Assert(id.Length != 0);

            XAttribute attr;
            return Root.DescendantNodes().OfType<XElement>().Where(
                    elm => (elm.Name.LocalName == "Style" && (attr = GetId(elm)) != null) && (attr.Value == id)
                ).First();
        }

        private Point _GetLocation(string text)
        {
            var coordinates = text.Split(new char[] { ',' });
            Debug.Assert(coordinates.Length == 3);

            var latitude = double.Parse(coordinates[0], NUMBER_FORMAT);
            var longitude = double.Parse(coordinates[1], NUMBER_FORMAT);

            return Point.Create(Engine, latitude, longitude);
        }

        private Point[] _GetPoints(string text)
        {
            var coordinates = text.Split(new char[] { ' ' });
            Debug.Assert(coordinates.Length > 1);


            var rv = new Point[coordinates.Length];
            for (var i = 0; i < rv.Length; ++i)
            {
                rv[i] = _GetLocation(coordinates[i]);
            }

            return rv;
        }

        

        public override string[] ContentIds
        {
            get
            {
                XAttribute attr;
                return Root.DescendantNodes().OfType<XElement>().Where(
                    elm => ((elm.Name.LocalName == "Placemark") && (attr = GetId(elm)) != null) && (attr.Value.Length != 0)
                ).Select(
                    elm => GetId(elm).Value
                ).ToArray();
            }
        }
        
        public override Point[] GetCoordinates(XElement node)
        {
            var id = GetId(node).Value;
            switch (node.Name.LocalName)
            {
                case "Placemark":
                    var point = _FindElement("Point", node);
                    if (point != null)
                    {
                        var coordinates = _FindElement("coordinates", point);
                        if (coordinates == null)
                        {
                            Debug.Print("Point {0} has no coordinates", id);
                            break;
                        }

                        var rv = _GetLocation(coordinates.Value);
                        if (rv.Latitude == 0.0 && rv.Longitude == 0.0)
                        {
                            Debug.Print("Point {0} has invalid coordinates", id);
                            break;
                        }

                        return new Point[] { rv };
                    }

                    var line = _FindElement("LineString", node);
                    if (line != null)
                    {
                        var coordinates = _FindElement("coordinates", line);
                        if (coordinates == null)
                        {
                            Debug.Print("LineString {0} has no coordinates", id);
                            break;
                        }

                        var rv = _GetPoints(coordinates.Value);
                        if (rv.Length == 0)
                        {
                            Debug.Print("LineString {0} has invalid coordinates", id);
                            break;
                        }

                        return rv;
                    }

                    if (point == null && line == null)
                    {
                        Debug.Print("Implemented only Point and Line.");
                    }
                    break;
            }

            return new Point[] { };
        }

        public override Label GetLabel(XElement node)
        {
            var rv = new Label();
            Debug.Assert(node != null);

            // Find the label element by his name.
            var name = _FindElement("name", node);
            if (name != null)
            {
                rv.Text = name.Value;
            }

            var description = _FindElement("description", node);
            rv.ShapeInfo = description != null ? description.Value : null;

            return rv;
        }

        public override Style GetStyle(XElement node)
        {
            var rv = new Style();
            Debug.Assert(node != null);

            // Find the style element by his id.
            var style_url = _FindElement("styleUrl", node);
            if (style_url == null)
            {
                Debug.Print("{0} has no style!", GetId(node).Value);
                return rv;
            }

            rv.Id = style_url.Value.Trim();
            if (rv.Id.StartsWith("#"))
                rv.Id = rv.Id.Substring(1);

            var style = _FindStyle(rv.Id);
            if (style == null)
            {
                Debug.Print("{0}/{1} style not found!", GetId(node).Value, rv.Id);
                return rv;
            }

            // Inflate the Icon information.
            var icon_style = _FindElement("IconStyle", style);
            if (icon_style != null)
            {
                var icon_style_icon = _FindElement("Icon", icon_style);
                if (icon_style_icon != null)
                {
                    var icon_style_icon_href = _FindElement("href", icon_style_icon);
                    if (icon_style_icon_href != null)
                    {
                        rv.IconPath = icon_style_icon_href.Value.Trim();
                        if (rv.IconPath.Length != 0 && !rv.IconPath.StartsWith(":/") && rv.IconPath.Substring(1, 2) != ":\\")
                            rv.IconPath = Engine.KmlPath.Substring(0, Engine.KmlPath.LastIndexOf('\\') + 1) + rv.IconPath.Replace('/', '\\');
                    }
                }
            }

            // Inflate the Line information.
            var line_style = _FindElement("LineStyle", style);
            if (line_style != null)
            {
                var line_style_color = _FindElement("color", line_style);
                if (line_style_color != null)
                    int.TryParse(line_style_color.Value, NumberStyles.HexNumber, NUMBER_FORMAT, out rv.LineColor);

                var line_style_width = _FindElement("width", line_style);
                if (line_style_width != null)
                    int.TryParse(line_style_width.Value, NumberStyles.Integer, NUMBER_FORMAT, out rv.LineWidth);
            }

            // Inflate the Label information.
            var label_style = _FindElement("LabelStyle", style);
            if (label_style != null)
            {
                var label_style_scale = _FindElement("scale", label_style);
                var label_style_face = _FindElement("face", label_style);

                rv.LabelFont = new StdFont()
                { 
                    Name = label_style_face == null ? "Arial" : label_style_face.Value,
                    Size = label_style_scale == null ? 8 : 8 * decimal.Parse(label_style_scale.Value, NUMBER_FORMAT)
                } as IFont;

                var label_style_color = _FindElement("color", label_style);
                if (label_style_color != null)
                    int.TryParse(label_style_color.Value, NumberStyles.HexNumber, NUMBER_FORMAT, out rv.LabelColor);

                var label_style_visible = _FindElement("visible", label_style);
                if ((label_style_visible != null) && !bool.TryParse(label_style_visible.Value, out rv.LabelVisible))
                    rv.LabelVisible = int.Parse(label_style_visible.Value) == 1;

                var label_style_style = _FindElement("style", label_style);
                if ((label_style_style != null))
                    int.TryParse(label_style_style.Value, out rv.LabelStyle);
            }

            return rv;
        }


        public override void SetStyle(XElement node, string style)
        {
            Debug.Assert((node != null) && (style.Length != 0));

            // Update the node style by changing his id. Update means always setup the style with node. 
            var style_url = _FindElement("styleUrl", node);
            if (style_url == null)
            {
                style_url = new XElement(XName.Get("styleUrl", node.Name.NamespaceName));
                node.Add(style_url);
            }
            style_url.Value = style;
        }

        
    }
}
