﻿namespace pipeline
{
    using GrymCore;
    using System;
    using System.Diagnostics;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Reflection;
    using System.Runtime.InteropServices;
    using System.Text;
    using System.Xml.Linq;
    using System.Windows.Forms;
    using Microsoft.Win32;    

    [Guid("6071a7e0-bba5-4d40-bc9e-a2831f6d43bd")]
    [ClassInterface(ClassInterfaceType.None)]
    [ComVisible(true)]
    public sealed class Plugin
        : IGrymPlugin
        , IGrymPluginInfo
        , IGrymPluginOptions
    {
        private static readonly int _DEFAULT_PORT = 5501;

        private static readonly string _REG_KEY = @"Software\stansage\Pipeline";
        private static readonly string _XML_INFO = @"
                                                    <grym_plugin_info>
                                                        <name>Pipeline v{0}</name>
                                                        <supported_languages>
                                                            <language>ru</language>
                                                        </supported_languages>
                                                        <requirements>
                                                            <requirement_api>API-1.4</requirement_api>
                                                        </requirements>
                                                    </grym_plugin_info>";

        private RegistryKey _regKey = Registry.CurrentUser.CreateSubKey(_REG_KEY);
        private JsonApi _api = new JsonApi();
        private IGrym _grym;
        private PluginEngine _engine;
        
        public Plugin()
        {
            
        }

        private void _Run()
        {
            Debug.Assert(_api != null && _engine != null);

            _engine.KmlPath = _regKey.GetValue("kmlpath", String.Empty) as string;
            if (_engine.KmlPath.Length == 0)
                return;

            if (_api == null)
                return;

            var port = (int)_regKey.GetValue("port", _DEFAULT_PORT);
            Debug.Print("Path {0}, Port {1}", _engine.KmlPath, port);
            //_tcpApi.Start(_engine, (int)port);
            _api.StartAsync(_engine, port);
        }


        public static Version Version
        {
            get
            {
                return Assembly.GetExecutingAssembly().GetName().Version;
            }
        }

        #region IGrymPlugin Members

        public void Initialize(GrymCore.Grym grym, GrymCore.IBaseViewThread baseViewThread)
        {
            try
            {
                _grym = grym;
                _engine = new PluginEngine(baseViewThread);
                _Run();
            }
            catch (Exception ex)
            {
                Debug.Fail(ex.Message, ex.StackTrace);
                Terminate();
                throw ex;
            }

        }

        public void Terminate()
        {
            if (_api != null)
                _api.Stop();

            _engine.Dispose();
            _regKey.Close();

            if (_grym != null)
                Marshal.FinalReleaseComObject(_grym);
            
            _grym = null;
            _engine = null;
            
            GC.SuppressFinalize(this);
            GC.Collect();
            GC.WaitForPendingFinalizers();
        }

        #endregion

        #region IGrymPluginInfo Members

        public string XMLInfo
        {
            get
            {
                return string.Format(_XML_INFO, Version);
            }
        }

        #endregion

        #region IGrymPluginOptions Members

        public bool OptionDialog()
        {
            var optionEditor = new OptionEditor();
            
            var port = (int)_regKey.GetValue("port", _DEFAULT_PORT);
            optionEditor.Port = port;

            _engine.KmlPath = _regKey.GetValue("kmlpath", String.Empty) as string;
            optionEditor.KmlPath = _engine.KmlPath;

            if (optionEditor.ShowDialog() != DialogResult.OK)
                return false;

            bool rv = false;

            if (optionEditor.KmlPath != _engine.KmlPath)
            {
                try
                {
                    var xml = XDocument.Load(optionEditor.KmlPath);
                    if (xml.Root == null)
                        throw new Exception("Invalid KML file");
                }
                catch (Exception ex)
                {
                    Debug.Fail(ex.Message, _engine.KmlPath);
                    _engine.KmlPath = string.Empty;
                    return false;
                }

                _engine.KmlPath = optionEditor.KmlPath;
                _regKey.SetValue("kmlpath", _engine.KmlPath);
                rv = true;
            }

            if (optionEditor.Port != port)
            {
                _regKey.SetValue("port", optionEditor.Port);
                rv = true;
            }

            if (rv)
            {
                if (_api != null)
                    _api.Stop();
                _Run();
            }

            return rv;
        }

        #endregion
    }
}
