﻿namespace pipeline
{
    using GrymCore;
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Drawing;
    using System.Linq;
    using System.Xml;
    using System.Xml.Linq;
    using System.Text;

    using Segment = System.Tuple<GrymCore.IMapPoint, GrymCore.IMapPoint>;

    internal sealed class Render
    {
        private enum Edge { Left, Top, Right, Bottom }

        private Dictionary<string, Image> _images = new Dictionary<string, Image>();
        private Segment[] _edges = new Segment[4];
        private Layer _layer;

        public Render(Layer layer)
        {
            Debug.Assert(layer != null);

            _layer = layer;
        }

        private IGrymObjectFactory _Factory
        {
            get
            {
                return _layer.Data.Engine.Factory;
            }

        }

        // Меняем значение переменных местами.
        private static void _Swap(ref double a, ref double b)
        {
            var tmp = a;
            a = b;
            b = tmp;
        }

        // Вычисление гипотенузы проекций отрезков.
        private static double _GetSegmentLength(Segment segment)
        {
            var dx = segment.Item2.X - segment.Item1.X;
            var dy = segment.Item2.Y - segment.Item1.Y;

            return Math.Sqrt(dx * dx + dy * dy);
        }

        // Площадь треугольника.
        private static double _GetArea(IMapPoint a, IMapPoint b, IMapPoint c)
        {
            return (b.X - a.X) * (c.Y - a.Y) - (b.Y - a.Y) * (c.X - a.X);
        }

        // Проверка пересечения проекции точки с проекцией отрезка.
        private static bool _IsIntersects(double a, double b, double c)
        {
            if (b > c)
                _Swap(ref b, ref c);

            return (a >= b) && (a <= c);
        }

        // Проверка пересечения проекций двух отрезков.
        private static bool _IsIntersects(double a, double b, double c, double d)
        {
            if (a > b)
                _Swap(ref a, ref b);
            if (c > d)
                _Swap(ref c, ref d);

            return Math.Max(a, c) <= Math.Min(b, d);
        }

        // Проверка пересечения двух отрезков (http://e-maxx.ru/algo/segments_intersection_checking).
        private static bool _IsIntersects(Segment a, Segment b)
        {
            return _IsIntersects(a.Item1.X, a.Item2.X, b.Item1.X, b.Item2.X)
                && _IsIntersects(a.Item1.Y, a.Item2.Y, b.Item1.Y, b.Item2.Y)
                && Math.Sign(_GetArea(a.Item1, a.Item2, b.Item1)) * Math.Sign(_GetArea(a.Item1, a.Item2, b.Item2)) <= 0
                && Math.Sign(_GetArea(b.Item1, b.Item2, a.Item1)) * Math.Sign(_GetArea(b.Item1, b.Item2, a.Item2)) <= 0;
        }

        // Проверка пересечения точки с отрезком.
        private static bool _IsIntersects(Segment segment, IMapPoint point)
        {
            Debug.Print("{0} [({1};{2}) - ({3};{4})] ({5};{6})", _GetArea(segment.Item1, point, segment.Item2),
                segment.Item1.X, segment.Item1.Y, segment.Item2.X, segment.Item2.Y, point.X, point.Y);
            return _GetArea(segment.Item1, point, segment.Item2) <= double.Epsilon;
        }

        private static IDevPoint _GetCenter(IDevRect rectangle)
        {
            IDevPoint rv = rectangle.Min;
            rv.X += rectangle.Width / 2;
            rv.Y += rectangle.Height / 2;

            return rv;
        }
        
        // Грани отрезка, построенные по углам  в порядке: Левый_Нижний Левый_Верхний Правый_Верхний Правый_Нижний.
        private static Segment _GetEdgeSegment(Edge edge, IMapPoint[] corners)
        {
            // 0 - 0 - 1 Left
            // 1 - 1 - 2 Top
            // 2 - 2 - 3 Right
            // 3 - 3 - 0 Bottom
            return Tuple.Create(corners[(int)edge], corners[edge == Edge.Bottom ? 0 : (int)edge + 1]);
        }

        // Простая оптимизация выборки картинок без какой-либо синхронизации.
        private Image _GetCachedImage(string path)
        {
            if (path == null || path.Length == 0)
                path = ":/marker"; // невидимое изображение

            if (!_images.ContainsKey(path))
            {
                var image = path.StartsWith(":/")
                          ? Properties.Resources.ResourceManager.GetObject(path.Substring(2), Properties.Resources.Culture) as Bitmap
                          : Bitmap.FromFile(path);

                _images.Add(path, image);
            }

            return _images[path];
        }

        // Углы произвольного прямоугольника в специальном порядке.
        private IMapPoint[] _GetCorners(IMapRect rectangle)
        {
            /*  -------------------
             *  | 1)  _    2) Max |
             *  | 0) Min   3)  _  |
             *  -------------------
             */

            return new IMapPoint[] {
                rectangle.Min,
                _Factory.CreateMapPoint(rectangle.Min.X, rectangle.Max.Y),
                rectangle.Max,
                _Factory.CreateMapPoint(rectangle.Max.X, rectangle.Min.Y)
            };
        }

        // Установка конечной точки отрезка на новое расстояние от начальной.
        private void _SetSegmentLength(ref Segment segment, double value)
        {
            var k = 1.0 - value / _GetSegmentLength(segment);
            
            var dx = segment.Item2.X - segment.Item1.X;
            var x = segment.Item2.X - k * dx;
            var dy = segment.Item2.Y - segment.Item1.Y;
            var y = segment.Item2.Y - k * dy;

            segment = Tuple.Create(segment.Item1, _Factory.CreateMapPoint(x, y));
        }

        // Поиск точки пересечения двух отрезков.
        private IMapPoint _GetCrossPoint(Segment a, Segment b)
        {
            var x = 0.0;
            var y = 0.0;
            var dxb = b.Item2.X - b.Item1.X;
            var dxa = a.Item2.X - a.Item1.X;

            if (Math.Abs(dxb) < double.Epsilon)
            {
                var k1 = (a.Item2.Y - a.Item1.Y) / dxb;
                var b1 = a.Item1.Y - k1 * a.Item1.X;

                x = b.Item1.X;
                y = k1 * x + b1;
            }
            else
            {

                var k2 = (b.Item2.Y - b.Item1.Y) / (b.Item2.X - b.Item1.X);
                var b2 = b.Item1.Y - k2 * b.Item1.X;

                if (Math.Abs(dxa) < double.Epsilon)
                {
                    x = a.Item1.X;
                    y = k2 * x + b2;

                }
                else
                {
                    var k1 = (a.Item2.Y - a.Item1.Y) / dxa;
                    var b1 = a.Item1.Y - k1 * a.Item1.X;

                    x = (b2 - b1) / (k1 - k2);
                    y = k1 * x + b1;
                }
            }

            return _Factory.CreateMapPoint(x, y);
        }

        // Определение координат для рисования элементов ХМЛ.
        private IMapPoint[] _GetDrawPoints(Data.Point[] coordinates, IMapRect rectangle)
        {
            // Строим граф из вершин со статусами: -1 - выход, +1 - вход, 0 - внутри видимого прямоугольника.
            var graph = new List<Tuple<IMapPoint, int>>(coordinates.Length);
            for (var i = 0; i < coordinates.Length; ++i)
            {
                var current = coordinates[i];
                var visible = current.InRect(rectangle);
                if (i == 0)
                {
                    // Первую точку пропускаем, а если она видимая, то добавляем ее в граф.
                    if (visible)
                        graph.Add(Tuple.Create(current.Local, 0));
                    continue;
                }

                // Отрезок строится по направлению от предыдущей точки к текущей.
                var segment = Tuple.Create(coordinates[i - 1].Local, current.Local);

                // Поиск пересекаемых граней с выстраиванием по порядку в зависимости от расстояния до начальной точки отрезка.
                // Отрезок может пересекать прямоугольник максимум в двух местах.
                // Принцип упорядочивания таков: 
                // в match.Item1 всегда будет номер ближней грани или -1, если нет пересечения с прмоугольником вообще,
                // а в match.Item2 будет номер дальней грани или -1, если отрезок пересекает только одну грань прямоугольника.
                var match = Tuple.Create(-1, -1, double.MaxValue);
                for (var j = 0; j < _edges.Length; ++j)
                {
                    if (!_IsIntersects(_edges[j], segment))
                        continue;

                    var d1 = _GetSegmentLength(Tuple.Create(_edges[j].Item1, segment.Item1));
                    var d2 = _GetSegmentLength(Tuple.Create(_edges[j].Item2, segment.Item1));
                    var distance = (d1 + d2) / 2.0;

                    if (distance < match.Item3)
                        match = Tuple.Create(j, match.Item1 == -1 ? -1 : match.Item1, distance);
                    else
                        match = Tuple.Create(match.Item1 == -1 ? j : match.Item1, j, distance);
                }

                if (match.Item1 != -1)
                {
                    Debug.Assert(match.Item1 < _edges.Length);

                    // Определяем статус новой вершины графа по предыдущей, или если граф пустой, по признаку нахожднения в области видимости.
                    var direction = graph.Count > 0 
                                  ? graph[graph.Count - 1].Item2 == 0 ? -1 : 1
                                  : visible ? -1 : 1;
                    graph.Add(Tuple.Create(_GetCrossPoint(_edges[match.Item1], segment), direction));
                }

                if (visible)
                {
                    //Debug.Assert(match.Item2 == -1);

                    // Вершина внутри видимой области.
                    graph.Add(Tuple.Create(current.Local, 0));
                }
                else if (match.Item2 != -1)
                {
                    Debug.Assert((match.Item2 < _edges.Length) && (match.Item1 != -1) && (graph.Count != 0) && (graph[graph.Count - 1].Item2 != 0));

                    // При пересечении с двумя гранями видимой области статус новой вершины противоположен статусу предыдущей вершины,
                    // причем предыдущая вершина не может находиться в видимой области по логике.
                    graph.Add(Tuple.Create(_GetCrossPoint(_edges[match.Item2], segment), -1 * graph[graph.Count - 1].Item2));
                }
                
            }

            // Вставляем разрывы на месте тех точек, у кого направление равно -1, т.е. после выхода из области видимости.
            var rv = graph.Select(x => x.Item1).ToList();
            for (int i = 0, offset = 0; i < graph.Count; ++i)
            {
                if (graph[i].Item2 == -1)
                {
                    rv.Insert(i + offset + (i == 0 ? 0 : 1), null);
                    ++offset;
                }
            }

            // Последовательность точек всегда должна начинаться с разрыва.
            if (rv.Count > 0 && rv[0] != null)
                rv.Insert(0, null);

            return rv.ToArray();
        }


        // Рисует ХМЛ элементы и подписи к ним с определенным стилем.
        public void Draw(IEnumerable<XElement> items, Painter painter, IMapDevice device, IMapRect window)
        {
            Debug.Assert(_Factory != null);
            
            // По углам строим грани области отрисовки.
            var corners = _GetCorners(window);
            for (int i = 0; i < _edges.Length; ++i)
                _edges[i] = _GetEdgeSegment((Edge)i, corners);
            
            foreach (var node in items)
            {
                var coordinates = _layer.Data.GetCoordinates(node);
                if (coordinates.Length == 0)
                    continue;
                
                var style = _layer.Data.GetStyle(node);

                // Условное обозначения это ХМЛ элемент с единственной координатой, иначе линия.
                if (coordinates.Length == 1)
                {
                    // Не показываем за границей видимого окна.
                    if (!coordinates[0].InRect(window))
                        continue;

                    // Рисуем изображение.
                    if (style.IconPath != null && style.IconPath.Length != 0)
                    {
                        var image = _GetCachedImage(style.IconPath);
                        foreach (var point in coordinates)
                        {
                            painter.Draw(device.MapToDevice(point.Local), image);
                        }
                    }
                }
                else
                {
                    var points = _GetDrawPoints(coordinates, window);
                    if (points.Length > 1)
                    {
                        // Рисуем ломаные.
                        painter.Pen.Color = Color.FromArgb(style.LineColor);
                        painter.Pen.Width = style.LineWidth;
                        painter.Pen.DashStyle = (System.Drawing.Drawing2D.DashStyle)style.LinePen;

                        List<IDevPoint> current = null;
                        var parts = new List<List<IDevPoint>>(points.Length / 2);
                        foreach (var pt in points)
                        {
                            if (pt != null)
                            {
                                Debug.Assert(current != null);
                                current.Add(device.MapToDevice(pt));
                            }
                            else if (current == null || current.Count > 0)
                            {
                                parts.Add(new List<IDevPoint>(points.Length));
                                current = parts[parts.Count - 1];
                            }
                        }
                        foreach (var part in parts)
                        {
                            if (part.Count > 1)
                                painter.Draw(part.ToArray());
                        }
                    }
                }

                // Вывод текстовые подписи по расчитанному среднему центру.
                var label = _layer.Data.GetLabel(node);
                if (style.LabelVisible && (label.Text != null) && (label.Text.Length != 0))
                {
                    var center = coordinates[0];
                    if (coordinates.Length > 1)
                    {
                        if (coordinates.Length % 2 == 1)
                        {
                            center = coordinates[coordinates.Length / 2];
                        }
                        else
                        {
                            var a = coordinates[coordinates.Length / 2 - 1];
                            var b = coordinates[coordinates.Length / 2];
                            center.Local = _Factory.CreateMapPoint((a.Local.X  + b.Local.X) / 2, (a.Local.Y  + b.Local.Y) / 2);
                        }
                    }

                    var font = Font.FromHfont((IntPtr)style.LabelFont.hFont);
                    painter.Draw(device.MapToDevice(center.Local), label.Text, font);
                }
            }
        }
    }
}
