﻿namespace pipeline
{
    using System;
    using System.Collections.Generic;
    using System.Diagnostics;
    using System.Linq;
    using System.Net;
    using System.Net.Sockets;
    using System.Text;
    using System.Xml.Linq;
    using System.IO;

    using AsyncArg = System.Tuple<System.Net.Sockets.TcpClient, byte[]>;

    internal sealed class TcpApi
    {
        private TcpListener _listener;
        private Layer _layer;

        public TcpApi()
        {
        }

        private void _OnAcceptClient(IAsyncResult result)
        {
            lock (this)
            {
                if (_listener == null)
                    return;
            }

            var client = _listener.EndAcceptTcpClient(result);
            Debug.Assert(client.GetStream() != null && _layer != null);

            Debug.Print("Incoming connection from {0}", client.Client.LocalEndPoint);

            var multiUser = result.AsyncState as TcpClient;
            if (multiUser != null && multiUser.Connected)
            {
                //TODO: Multi user support
                Debug.Print("Already connected with {0}", multiUser.Client.LocalEndPoint);
                client.GetStream().WriteByte(0);
                client.Close();
                client = multiUser;
            }
            else
            {
                Debug.Print("Accepted connection from {0} at {1}", client.Client.LocalEndPoint, DateTime.UtcNow.ToString(@"dd.MM.yyyy a\t HH:mm:ss.f"));

                using (var tmp = File.Open(_layer.Data.Engine.KmlPath, FileMode.Open))
                {
                    tmp.CopyTo(client.GetStream(), client.SendBufferSize);
                }

                var buf = new byte[1024];
                client.GetStream().BeginRead(buf, 0, buf.Length, new AsyncCallback(_OnClientRead), new AsyncArg(client, buf));
                Debug.Print("Xml sent to {0}", client.Client.LocalEndPoint);
            }
            
            _listener.BeginAcceptTcpClient(new AsyncCallback(_OnAcceptClient), client);

            
        }

        
        private void _OnClientRead(IAsyncResult result)
        {
            var arg = (result.AsyncState as AsyncArg);
            Debug.Assert(arg != null && arg.Item1 != null && arg.Item2 != null);

            var data = arg.Item2;
            var stream = arg.Item1.GetStream();
            Debug.Assert(stream != null);
            
            try
            {
                var count = stream.EndRead(result);
                data = data.Take(count).ToArray();

                if (data.Length == 0)
                    throw new Exception("client is dead");
            }
            catch (Exception ex)
            {
                Debug.Print("Lost connection from {0}: (1)\n{2}", arg.Item1.Client.LocalEndPoint, ex.Message, ex.StackTrace);
                stream.Close();
                return;
            }

            var tail = data.SkipWhile(x => x != '\n').Skip(1).ToArray();
            if (tail.Length == data.Length)
            {
                Debug.Fail("Request is too big", Encoding.Default.GetString(tail));
                arg.Item1.Close();
                return;
            }

            data = data.Take(data.Length - tail.Length).ToArray();
            var request = Encoding.Default.GetString(data).Trim();
            var response = "OK";

            try
            {
                if (request.Length == 0)
                    throw new Exception("Empty request");


                var pair = request.Split(new char[] { ' ' });
                if (pair.Length != 2)
                    throw new Exception("To small arguments");

                if (pair.Length > 2)
                {
                    Debug.Print("To many arguments [{0}]", string.Join(":", pair));
                    pair = pair.Take(2).ToArray();
                }

                _layer.Update(pair.First(), pair.Last());
            }
            catch (Exception ex)
            {
                Debug.Print("{0}\n{1}", ex.Message, ex.StackTrace);
                response = ex.Message + " - " + ex.StackTrace;
            }

            try
            {
                var writer = new StreamWriter(arg.Item1.GetStream());
                writer.WriteLine(response);
                writer.Flush();

                // Copy a tail to the head of next packet
                for (var i = 0; i < tail.Length; ++i)
                    arg.Item2[i] = tail[i];
            }
            catch (Exception ex)
            {
                Debug.Print("{0}\n{1}", ex.Message, ex.StackTrace);
                return;
            }

            stream.BeginRead(arg.Item2, tail.Length, arg.Item2.Length - tail.Length, new AsyncCallback(_OnClientRead), arg);
        }

        public void Start(PluginEngine engine, int port)
        {
            _layer = new Layer(engine);

            _listener = new TcpListener(IPAddress.Any, port);
            _listener.Start();
            _listener.BeginAcceptTcpClient(new AsyncCallback(_OnAcceptClient), null);

            Debug.Print("Session started");
        }

        public void Stop()
        {
            lock (this)
            {
                if (_listener != null)
                    _listener.Stop();
                _listener = null;
            }

            if (_layer != null)
            {
                _layer.Dispose();
                _layer = null;
            }
            
            GC.SuppressFinalize(this);
            GC.Collect();
            GC.WaitForPendingFinalizers();

            Debug.Print("Session stopped");
        }
    }
}
